"""Async iterable that wraps an EventEmitter."""

import asyncio
import collections


class EventIterator:

    """An iterator who values are the payloads from emitted events."""

    def __init__(self, emitter, event):
        """Initialize the iterator with an emitter and event to fire on."""
        self._emitter = emitter
        self._event = event
        self._emitter.on(event, self._push)
        self._data = collections.deque()
        self._future = None

    async def _push(self, *args, **kwargs):
        """Push new data into the buffer. Resume looping if paused."""
        self._data.append((args, kwargs))
        if self._future is not None:

            future, self._future = self._future, None
            future.set_result(True)

    async def __anext__(self):
        """Fetch the next set of values. Wait for new values if empty."""
        if self._data:

            return self._data.popleft()

        self._future = asyncio.Future()
        await self._future
        return self._data.popleft()


class EventIterable:

    """An iterable object the iterator for which loops on each fired event."""

    def __init__(self, emitter, event):
        """Initialize the iterable with an emitter and target event."""
        self._emitter = emitter
        self._event = event

    async def __aiter__(self):
        """Get a new EventIterator object."""
        return EventIterator(self._emitter, self._event)



"""Event emitter mixin class."""

import asyncio
import collections
import contextlib
import functools
import itertools
import warnings


async def _try_catch_coro(emitter, event, listener, coro):
    """Coroutine wrapper to catch errors after async scheduling.

    Args:
        emitter (EventEmitter): The event emitter that is attempting to
            call a listener.
        event (str): The event that triggered the emitter.
        listener (async def): The async def that was used to generate the coro.
        coro (coroutine): The coroutine that should be tried.

    If an exception is caught the function will use the emitter to emit the
    failure event. If, however, the current event _is_ the failure event then
    the method reraises. The reraised exception may show in debug mode for the
    event loop but is otherwise silently dropped.
    """
    try:

        await coro

    except Exception as exc:

        if event == emitter.LISTENER_ERROR_EVENT:

            raise

        emitter.emit(emitter.LISTENER_ERROR_EVENT, event, listener, exc)


class EventEmitter:

    """A mixin that provides methods for publishing and listening to events.

    Attributes:
        DEFAULT_MAX_LISTENERS (int): The default number of listeners per
            event before the object begins emitting warnings. This value can
            be adjust by subclassing and setting this value. Alternatively, it
            can be adjusted per instance by using the max_listeners property.
            The default value is 10.
        LISTENER_ERROR_EVENT (str): The name of the event that should be
            listened to in the event of an exception within a listener.
    """

    DEFAULT_MAX_LISTENERS = 10
    LISTENER_ERROR_EVENT = 'listener-error'

    def __init__(self, loop=None):
        """Initialize the emitter with an event loop."""
        self._loop = loop or asyncio.get_event_loop()
        self._listeners = collections.defaultdict(list)
        self._once = collections.defaultdict(list)
        self._max_listeners = self.DEFAULT_MAX_LISTENERS

    def _check_limit(self, event):
        """Check if the listener limit is hit and warn if needed."""
        if self.count(event) > self.max_listeners:

            warnings.warn(
                'Too many listeners for event {}'.format(event),
                ResourceWarning,
            )

    def add_listener(self, event, listener):
        """Bind a listener to a particular event.

        Args:
            event (str): The name of the event to listen for. This may be any
                string value.
            listener (def or async def): The callback to execute when the event
                fires. This may be a sync or async function.
        """
        self.emit('new_listener', event, listener)
        self._listeners[event].append(listener)
        self._check_limit(event)
        return self

    on = add_listener

    def once(self, event, listener):
        """Add a listener that is only called once."""
        self.emit('new_listener', event, listener)
        self._once[event].append(listener)
        self._check_limit(event)
        return self

    def remove_listener(self, event, listener):
        """Remove a listener from the emitter.

        Args:
            event (str): The event name on which the listener is bound.
            listener: A reference to the same object given to add_listener.

        Returns:
            bool: True if a listener was removed else False.

        This method only removes one listener at a time. If a listener is
        attached multiple times then this method must be called repeatedly.
        Additionally, this method removes listeners first from the those
        registered with 'on' or 'add_listener'. If none are found it continue
        to remove afterwards from those added with 'once'.
        """
        with contextlib.suppress(ValueError):

            self._listeners[event].remove(listener)
            return True

        with contextlib.suppress(ValueError):

            self._once[event].remove(listener)
            return True

        return False

    @property
    def max_listeners(self):
        """Get the max number of listeners before warning."""
        return self._max_listeners

    @max_listeners.setter
    def max_listeners(self, value):
        """Set the max number of listeners before warning."""
        self._max_listeners = value

    def listeners(self, event):
        """Get an iterable of all listeners for the given event.

        Args:
            event (str): The name of the event for which to generate an
                iterable of listeners.

        The resulting iterable contains all listeners regardless of whether
        they were registered with 'on'/'add_listener' or 'once'.
        """
        return self._listeners[event][:] + self._once[event][:]

    def _dispatch_coroutine(self, event, listener, *args, **kwargs):
        """Schedule a coroutine for execution.

        Args:
            event (str): The name of the event that triggered this call.
            listener (async def): The async def that needs to be executed.
            *args: Any number of positional arguments.
            **kwargs: Any number of keyword arguments.

        The values of *args and **kwargs are passed, unaltered, to the async
        def when generating the coro. If there is an exception generating the
        coro, such as the wrong number of arguments, the emitter's error event
        is triggered. If the triggering event _is_ the emitter's error event
        then the exception is reraised. The reraised exception may show in
        debug mode for the event loop but is otherwise silently dropped.
        """
        try:

            coro = listener(*args, **kwargs)

        except Exception as exc:

            if event == self.LISTENER_ERROR_EVENT:

                raise

            return self.emit(self.LISTENER_ERROR_EVENT, event, listener, exc)

        asyncio.ensure_future(
            _try_catch_coro(self, event, listener, coro),
            loop=self._loop,
        )

    def _dispatch_function(self, event, listener, *args, **kwargs):
        """Execute a sync function.

        Args:
            event (str): The name of the event that triggered this call.
            listener (def): The def that needs to be executed.
            *args: Any number of positional arguments.
            **kwargs: Any number of keyword arguments.

        The values of *args and **kwargs are passed, unaltered, to the def
        when exceuting. If there is an exception executing the def, such as the
        wrong number of arguments, the emitter's error event is triggered. If
        the triggering event _is_ the emitter's error event then the exception
        is reraised. The reraised exception may show in debug mode for the
        event loop but is otherwise silently dropped.
        """
        try:

            return listener(*args, **kwargs)

        except Exception as exc:

            if event == self.LISTENER_ERROR_EVENT:

                raise

            return self.emit(self.LISTENER_ERROR_EVENT, event, listener, exc)

    def _dispatch(self, event, listener, *args, **kwargs):
        """Dispatch an event to a listener.

        Args:
            event (str): The name of the event that triggered this call.
            listener (def or async def): The listener to trigger.
            *args: Any number of positional arguments.
            **kwargs: Any number of keyword arguments.

        This method inspects the listener. If it is a def it dispatches the
        listener to a method that will execute that def. If it is an async def
        it dispatches it to a method that will schedule the resulting coro with
        the event loop.
        """
        if (
            asyncio.iscoroutinefunction(listener) or
            isinstance(listener, functools.partial) and
            asyncio.iscoroutinefunction(listener.func)
        ):

            return self._dispatch_coroutine(event, listener, *args, **kwargs)

        return self._dispatch_function(event, listener, *args, **kwargs)

    def emit(self, event, *args, **kwargs):
        """Call each listener for the event with the given arguments.

        Args:
            event (str): The event to trigger listeners on.
            *args: Any number of positional arguments.
            **kwargs: Any number of keyword arguments.

        This method passes all arguments other than the event name directly
        to the listeners. If a listener raises an exception for any reason the
        'listener-error', or current value of LISTENER_ERROR_EVENT, is emitted.
        Listeners to this event are given the event name, listener object, and
        the exception raised. If an error listener fails it does so silently.

        All event listeners are fired in a deferred way so this method returns
        immediately. The calling coro must yield at some point for the event
        to propagate to the listeners.
        """
        listeners = self._listeners[event]
        listeners = itertools.chain(listeners, self._once[event])
        self._once[event] = []
        for listener in listeners:

            self._loop.call_soon(
                functools.partial(
                    self._dispatch,
                    event,
                    listener,
                    *args,
                    **kwargs,
                )
            )

        return self

    def count(self, event):
        """Get the number of listeners for the event.

        Args:
            event (str): The event for which to count all listeners.

        The resulting count is a combination of listeners added using
        'on'/'add_listener' and 'once'.
        """
        return len(self._listeners[event]) + len(self._once[event])



"""Test suites for the EventIterable implementation."""

import asyncio
import functools

from . import emitter
from . import iterable


def async_test(loop=None):
    """Wrap an async test in a run_until_complete for the event loop."""
    loop = loop or asyncio.get_event_loop()

    def _outer_async_wrapper(func):
        """Closure for capturing the configurable loop."""
        @functools.wraps(func)
        def _inner_async_wrapper(*args, **kwargs):

            return loop.run_until_complete(func(*args, **kwargs))

        return _inner_async_wrapper

    return _outer_async_wrapper


@async_test()
async def test_iter_pauses_until_ready():
    """Check that the EventIterable waits until data available."""
    e = emitter.EventEmitter()
    i = iterable.EventIterable(e, 'test')
    sentinel = object()
    asyncio.get_event_loop().call_later(
        .1,
        functools.partial(e.emit, 'test', sentinel, test=sentinel),
    )
    async for args, kwargs in i:

        assert sentinel in args
        assert 'test' in kwargs
        assert kwargs['test'] is sentinel
        break


@async_test()
async def test_iter_pulls_from_buffer():
    """Check that the EventIterable emits buffered data if available."""
    e = emitter.EventEmitter()
    i = iterable.EventIterable(e, 'test')
    iterator = await i.__aiter__()
    sentinel = object()
    e.emit('test', sentinel, test=sentinel)
    await asyncio.sleep(0)
    args, kwargs = await iterator.__anext__()
    assert sentinel in args
    assert 'test' in kwargs
    assert kwargs['test'] is sentinel



"""Setuptools configuration for eventemitter."""

from setuptools import setup
from setuptools import find_packages


with open('README.rst', 'r') as readmefile:

    README = readmefile.read()

setup(
    name='eventemitter',
    version='0.1.1',
    url='https://github.com/asyncdef/eventemitter',
    description='Tools for working with async events.',
    author="Kevin Conway",
    author_email="kevinjacobconway@gmail.com",
    long_description=README,
    license='Apache 2.0',
    packages=find_packages(exclude=['build', 'dist', 'docs']),
    install_requires=[

    ],
    extras_require={
        'testing': [
            "pep257",
            "pep8",
            "pyenchant",
            "pyflakes",
            "pylint",
            "pytest",
            "pytest-cov",
        ],
    },
    entry_points={
        'console_scripts': [

        ],
    },
    include_package_data=True,
)



"""Test suites for the EventEmitter implementation."""

import asyncio
import functools
import warnings

from . import emitter


def async_test(loop=None):
    """Wrap an async test in a run_until_complete for the event loop."""
    loop = loop or asyncio.get_event_loop()

    def _outer_async_wrapper(func):
        """Closure for capturing the configurable loop."""
        @functools.wraps(func)
        def _inner_async_wrapper(*args, **kwargs):

            return loop.run_until_complete(func(*args, **kwargs))

        return _inner_async_wrapper

    return _outer_async_wrapper


@async_test()
async def test_emitter_emits_async_coro():
    """Check that listeners are called async when given async def."""
    e = emitter.EventEmitter()
    value = {"touched": False}

    async def touch_value():
        """Flip the test bit to True."""
        value['touched'] = True

    e.on('test', touch_value)
    e.emit('test')
    assert value['touched'] is False
    await asyncio.sleep(0)
    assert value['touched'] is True


@async_test()
async def test_emitter_emits_async_func():
    """Check that listeners are called async when given def."""
    e = emitter.EventEmitter()
    value = {"touched": False}

    def touch_value():
        """Flip the test bit to True."""
        value['touched'] = True

    e.on('test', touch_value)
    e.emit('test')
    assert value['touched'] is False
    await asyncio.sleep(0)
    assert value['touched'] is True


@async_test()
async def test_emitter_calls_multiple_times():
    """Check that listeners are called on each fired event."""
    e = emitter.EventEmitter()
    value = {'count': 0}

    def inc_value():
        """Up the counter."""
        value['count'] += 1

    e.on('test', inc_value)
    e.emit('test')
    e.emit('test')
    e.emit('test')
    await asyncio.sleep(0)
    assert value['count'] == 3


@async_test()
async def test_emitter_calls_only_once():
    """Check that once listeners are not called multiple times."""
    e = emitter.EventEmitter()
    value = {'count': 0}

    def inc_value():
        """Up the counter."""
        value['count'] += 1

    e.once('test', inc_value)
    e.emit('test')
    e.emit('test')
    e.emit('test')
    await asyncio.sleep(0)
    assert value['count'] == 1


def test_remove_listener_empty():
    """Check that False is returned when no listener is removed."""
    e = emitter.EventEmitter()
    sentinel = object()
    assert e.remove_listener('test', sentinel) is False


def test_remove_listener_full():
    """Check that True is returned when a listener is removed."""
    e = emitter.EventEmitter()
    sentinel = object()
    e.on('test', sentinel)
    assert e.remove_listener('test', sentinel) is True
    assert e.remove_listener('test', sentinel) is False


def test_remove_listener_once():
    """Check that True is returned when a once listener is removed."""
    e = emitter.EventEmitter()
    sentinel = object()
    e.once('test', sentinel)
    assert e.remove_listener('test', sentinel) is True
    assert e.remove_listener('test', sentinel) is False


def test_count_shows_all_listeners():
    """Check that count returns the number of all listeners for an event."""
    e = emitter.EventEmitter()
    sentinel = object()
    e.on('test', sentinel)
    e.on('test', sentinel)
    e.once('test', sentinel)
    assert e.count('test') == 3


def test_listers_give_all():
    """Check that listeners() returns all registered listeners for event."""
    e = emitter.EventEmitter()
    sentinel = object()
    e.on('test', sentinel)
    e.on('test', sentinel)
    e.once('test', sentinel)
    listeners = e.listeners('test')
    assert len(listeners) == 3
    for listener in listeners:

        assert listener is sentinel


def test_listener_limit_warning():
    """Check that a warning is emitted if too many listeners are added."""
    e = emitter.EventEmitter()
    e.max_listeners = 1
    e.on('test', None)
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        e.on('test', None)

    assert len(w) == 1
    assert issubclass(w[-1].category, ResourceWarning)
    assert "test" in str(w[-1].message)


@async_test()
async def test_listener_error_event_coro():
    """Check that the LISTENER_ERROR_EVENT is fired on async exceptions."""
    e = emitter.EventEmitter()

    async def fail(*args, **kwargs):
        """Raise an exception."""
        raise TypeError()

    caught = {'value': False}

    async def catch(event, listener, exc):
        """Catch the raised exception and assert it."""
        assert event == 'test'
        assert listener is fail
        assert isinstance(exc, TypeError)
        caught['value'] = True

    e.on('test', fail)
    e.on(e.LISTENER_ERROR_EVENT, catch)
    e.emit('test')
    while not caught['value']:

        await asyncio.sleep(0)


@async_test()
async def test_listener_error_event_coro_create():
    """Check that the LISTENER_ERROR_EVENT is fired when coro create fails."""
    e = emitter.EventEmitter()

    async def fail():
        """Rais TypeError if given any arguments."""
        return None

    caught = {'value': False}

    async def catch(event, listener, exc):
        """Catch the raised exception and assert it."""
        assert event == 'test'
        assert listener is fail
        assert isinstance(exc, TypeError)
        caught['value'] = True

    e.on('test', fail)
    e.on(e.LISTENER_ERROR_EVENT, catch)
    e.emit('test', 1, 2, 3)
    while not caught['value']:

        await asyncio.sleep(0)


@async_test()
async def test_listener_error_event_func():
    """Check that the LISTENER_ERROR_EVENT is fired on sync exceptions."""
    e = emitter.EventEmitter()

    def fail(*args, **kwargs):
        """Raise an exception."""
        raise TypeError()

    caught = {'value': False}

    def catch(event, listener, exc):
        """Catch the raised exception and assert it."""
        assert event == 'test'
        assert listener is fail
        assert isinstance(exc, TypeError)
        caught['value'] = True

    e.on('test', fail)
    e.on(e.LISTENER_ERROR_EVENT, catch)
    e.emit('test')
    while not caught['value']:

        await asyncio.sleep(0)
